#ifndef SGG_H
#define SGG_H

//This guy always wants to be at the very top of everything whenever 
//you are using STL with MSVC
#pragma warning(disable: 4786) 

#include <string>
#include <vector>
#include <functional>
#include <algorithm>

#include <pa.h>
#include <sven.h>

namespace sgg {

using namespace sven;
using namespace std_ex;

typedef void (*void_fun)();
typedef int (*int_fun)();

//the MSVC mem_fun implementation doesn't seem to work if it is given a 
//void member function (which is terrible if you are tring to use for_each).
//So I have made my own mem_fun struct definitions for the void functions
//of my 3 top level Registerable abstract classes.

struct Registerable {
	virtual ~Registerable() {}
};

//So a drawable is anything that can be drawn.
//Drawables are generally expected to be objects in a screen,
//rather than Screens themselves.  Thus,
//even though all Screens have a draw function, 
//not all Screens are Drawables.
struct Drawable : public Registerable {
  struct mem_draw { void operator()(Drawable * d) { d->draw(); }};
  virtual void draw()=0;
};

//A Pickable is an object in a screen that can be picked
//when the screen's pick() function get's called.
//Notice that pick takes no arguments, and thus a GUI object
//with several distinct pickable components cannot be well 
//represented as a single Pickable.
//For something like a list I suggest implmenting it using the 
//MuliPick abstraction.
struct Pickable : public Registerable {
  struct mem_pick { void operator()(Pickable * p) { p->pick(); }};
  struct mem_p_draw { void operator()(Pickable * p) { p->p_draw(); }};

  virtual void p_draw()=0;
  virtual void pick()=0;
};


//things that you might want to do in a refresh:
//update particle engine or other constantly running effects
//handle keyboard input
//update dynamic elements of the screen (say you have planets that slowly
//                    orbit their star).
struct Refreshable : public Registerable {
	struct mem_refresh { void operator()(Refreshable * r) { r->refresh();}};

	virtual void refresh()=0;
};


struct KillFlag : public Registerable {
	bool killme;
	KillFlag() : killme(false) {}
};



//There is a stack of screens, 'g_screens'.  
//The main program loop works as follows:
//Run the draw function of the screen ontop of the stack.
//Run the refresh function of the screen ontop of the stack.
//Thus the following abstract class definition:
struct Screen {
  virtual void refresh() =0;
  virtual ~Screen() {}
};

//A Pick_Behavior specifies what a given Pickable does when
//its pick function gets called.  It is handy to be able to
//separate the pick behavior out from the rest of the Pickable
//object, as a Pickables p_draw() and draw() functions can often
//be fully specified just by reading in external data.
//A Pick_Behavior is abstract class for a void function object.
//Sadly, this forces us to use pointers to function objects,
//which destroys the compiler's ability to properly
//optimize them, and generally makes the code look a little less
//pretty than it might otherwise.
struct Pick_Behavior {
  virtual void operator()()=0;
};


namespace gui_error {
	struct screen_stack_empty {};
};

extern bool g_close_screen;
extern bool g_quit_flag;
extern vector<Screen *> g_screens;

//main is expected to contain a loop of roughly the form:
//'while(!g_quit) gui_update();'
inline void gui_update() {
	if(!g_screens.size()) {
		//throw gui_error::screen_stack_empty();
		g_quit_flag=true;
		return;
	}
    if(g_close_screen) {
      if(g_screens.size()>1) {
		Screen * kill = g_screens.back();
		g_screens.pop_back();
		delete kill;
		g_close_screen=false;
      }
      else {
		  throw gui_error::screen_stack_empty();
		  return;
      }
    }
    g_screens.back()->refresh();
}


}

#endif